from multiprocessing import Pool
from typing import Optional

import typer
from rich import print

from gdshoplib import Platform, Product
from gdshoplib.packages.cache import KeyDBCache
from gdshoplib.packages.feed import Feed
from gdshoplib.packages.s3 import S3, SimpleS3Data
from gdshoplib.services.notion.database import Database
from gdshoplib.services.notion.notion import Notion
from gdshoplib.services.vk.market import VKMarket
from gdshoplib.services.vk.vk import VK

app = typer.Typer()


@app.command()
def sku_set():
    for page in Product.query(
        params={
            "filter": {
                "and": [
                    {"property": "Наш SKU", "rich_text": {"is_empty": True}},
                    {"property": "Цена (eur)", "number": {"is_not_empty": True}},
                ]
            }
        },
    ):
        sku = page.generate_sku()
        while not Product.query(filter={"sku": sku}):
            sku = page.generate_sku()

        page.notion.update_prop(
            page.id, params={"properties": {"Наш SKU": [{"text": {"content": sku}}]}}
        )
        print(Product(page.id).sku)


def price_update_action(id):
    product = Product(id)
    props_map = {
        "Текущая Цена": product.price.now,
        "Цена комплекта": product.price.get_kit_price(),
        "Безубыточность": product.price.neitral,
        "Текущая Скидка": product.price.current_discount,
        "Агентская Цена": product.price.neitral,
        "Агентский комплект": product.price.get_kit_price(base_price="neitral"),
        "Себестоимость": product.price.gross,
    }

    for k, v in props_map.items():
        product.notion.update_prop(
            product.id, params={"properties": {k: {"number": v}}}
        )

    print(product.sku)


@app.command()
def price_update(
    sku: Optional[str] = typer.Option(None), single: bool = typer.Option(False)
):
    if sku:
        price_update_action(Product.get(sku).id)
        return

    if single:
        for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
            price_update_action(product["id"])
    else:
        with Pool(3) as p:
            for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                p.apply_async(price_update_action, (product["id"],))
            p.close()
            p.join()


def generate_description(id):
    product = Product(id)
    product.description.warm_description_blocks()
    for platform, block in product.description.description_blocks.items():
        key = platform.split(":")[-1]
        platform = Platform.get_platform(key=key)
        new_description = product.description.generate(platform.manager)
        Notion().update_block(
            block.id,
            params={"code": {"rich_text": [{"text": {"content": new_description}}]}},
        )
        print(f"{product.sku}: {platform}")


@app.command()
def description_regenerate(single: bool = typer.Option(False)):
    if single:
        for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
            generate_description(product["id"])
    else:
        with Pool(3) as p:
            for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                p.apply_async(generate_description, (product["id"],))
            p.close()
            p.join()


def description_check_action(id):
    product = Product(id)
    for platform_manager in Platform.__subclasses__():
        block = product.description.get_description_block(
            platform_key=platform_manager.KEY
        )
        print(f'{product.sku} {platform_manager}: {block.check if block else "None"}')


@app.command()
def description_check(single: bool = typer.Option(False)):
    with Pool(3) as p:
        for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
            p.apply_async(description_check_action, (product["id"],))
        p.close()
        p.join()


@app.command()
def cache_clean(id: Optional[str] = typer.Option(None)):
    # TODO: сделать удаление по ID
    KeyDBCache().clean(r"[blocks|pages|databases]*")


def cache_warm_func(id):
    product = Product(id)
    product.price.now
    product.kit
    product.notes
    product.specifications
    product.tags
    product.media
    product.description
    product.brand.title
    product.description.warm_description_blocks()
    print(f"{product.sku}: {product.last_edited_time}")


@app.command()
def cache_warm(
    only_exists: bool = typer.Option(False),
    single: bool = typer.Option(False),
    only_edited: bool = typer.Option(True),
    sku: Optional[str] = typer.Option(None),
):
    if sku:
        cache_warm_func(Product.get(sku).id)
        return

    if single:
        with Database(
            Product.SETTINGS.PRODUCT_DB, notion=Notion(caching=True)
        ) as database:
            params = {}
            if only_edited and database.get_update_time():
                print(f"Фильтрация от даты: {database.get_update_time()}")
                params = database.edited_filter()

            for product in database.pages(params=params):
                skipped = False
                if only_exists:
                    if KeyDBCache().exists(product["id"]):
                        print(f"{product['id']}: SKIPPED")
                        skipped = True

                if not skipped:
                    cache_warm_func(product["id"])
    else:
        with Pool(3) as p:
            with Database(
                Product.SETTINGS.PRODUCT_DB, notion=Notion(caching=True)
            ) as database:
                params = {}
                if only_edited and database.get_update_time():
                    print(f"Фильтрация от даты: {database.get_update_time()}")
                    params = database.edited_filter()

                for product in database.pages(params=params):
                    skipped = False
                    if only_exists:
                        if KeyDBCache().exists(product["id"]):
                            print(f"{product['id']}: SKIPPED")
                            skipped = True

                    if not skipped:
                        p.apply_async(cache_warm_func, (product["id"],))
            p.close()
            p.join()


@app.command()
def cache_count():
    print(KeyDBCache().count())


def cache_check_action(id):
    for block in Notion().get_blocks(id):
        exists = KeyDBCache().exists(block["id"])
        print(f"{block['id']}: {exists}")


@app.command()
def cache_check(single: bool = typer.Option(False)):
    if single:
        for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
            cache_check_action(product["id"])
    else:
        with Pool(3) as p:
            for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                p.apply_async(cache_check_action, (product["id"],))
            p.close()
            p.join()


def warm_product_media(id, with_badges):
    for media in Product(id).media:
        media.fetch()
        print(f"{media.s3}: {media.exists()}")
        if with_badges and media.badges:
            render = media.apply_badges()
            badged = S3(
                SimpleS3Data(
                    render.content,
                    None,
                    render.info["mime"],
                    file_info={
                        "id": media.id,
                        "format": render.info["format"],
                        "prefix": "BADED",
                    },
                    parent=media.parent,
                )
            )
            badged.put()
            print(f"{badged} {badged.exists()}")


@app.command()
def media_warm(
    single: bool = typer.Option(False), with_badges: bool = typer.Option(False)
):
    if single:
        for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
            warm_product_media(product["id"], with_badges)
    else:
        with Pool(7) as p:
            for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                p.apply_async(warm_product_media, (product["id"], with_badges))
            p.close()
            p.join()


@app.command()
def media_clean():
    s3 = S3(SimpleS3Data(None, file_key=None))
    s3.clean(pattern="Contents[? !contains(Key, `feed.`)][]")


@app.command()
def media_search(file_key=None):
    s3 = S3(SimpleS3Data(None, file_key=None))
    for object in s3.search(pattern=file_key):
        print(object)


@app.command()
def media_count():
    count = 0
    for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
        count += len(Product(product["id"]).media)

    print(count)


def media_check_action(id):
    accepted_formats = (
        "png",
        "jpg",
        "jpeg",
    )
    product = Product(id)
    for media in Product(product.id).media:
        accepted = media.format in accepted_formats
        print(
            f"{media.file_key}: {media.exists()} {'ACCEPTED' if accepted else 'REJECTED'}"
        )


@app.command()
def media_check(single: bool = typer.Option(False)):
    if single:
        for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
            media_check_action(product["id"])
    else:
        with Pool(3) as p:
            for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                p.apply_async(media_check_action, (product["id"],))
            p.close()
            p.join()


def warm_platfrom_feed(key):
    platform = Feed.get_platform_class(key=key)
    platform().push_feed()
    print(platform)


@app.command()
def feed_warm(platform_key=None):
    if platform_key:
        warm_platfrom_feed(platform_key)
        return

    for platform in [Feed, *[feed for feed in Feed.__subclasses__()]]:
        warm_platfrom_feed(platform.KEY)


@app.command()
def vk_get_access_code(code=None):
    if not code:
        VK().get_oauth_code()
        code = typer.prompt("Код")

    print(VK().get_access_token(code))


@app.command()
def vk_health():
    assert VKMarket().list(), "Запрос в VK не выполняется"
    print("OK")


if __name__ == "__main__":
    app()



from typing import Optional

import boto3
from botocore.exceptions import ClientError
from pydantic import BaseModel

from gdshoplib.core.settings import S3Settings


class S3:
    def __init__(self, data):
        self.data = data
        self.content = None
        self.s3_settings = S3Settings()
        self.__session = None
        self.__client = None
        self.__iter = None

    @property
    def session(self):
        if not self.__session:
            self.__session = boto3.session.Session()

        return self.__session

    @property
    def s3(self):
        if not self.__client:
            self.__client = self.session.client(
                service_name="s3",
                endpoint_url=self.s3_settings.ENDPOINT_URL,
                aws_access_key_id=self.s3_settings.ACCESS_KEY,
                aws_secret_access_key=self.s3_settings.SECRET_KEY,
            )
        return self.__client

    def put(self):
        return self.s3.put_object(
            Bucket=self.s3_settings.BUCKET_NAME,
            Key=self.data.file_key,
            Body=self.data.content,
            ACL="public-read",
            StorageClass="ICE",
            ContentType=self.data.mime,
        )

    def exists(self, key=None):
        try:
            self.s3.head_object(
                Bucket=self.s3_settings.BUCKET_NAME, Key=key or self.data.file_key
            )
        except ClientError:
            return False
        else:
            return True

    def __iter__(self):
        paginator = self.s3.get_paginator("list_objects_v2")
        iterator = paginator.paginate(
            Bucket=self.s3_settings.BUCKET_NAME, PaginationConfig={"PageSize": 100}
        )
        self.__iter = iter([*(page["Contents"] for page in iterator)][0])
        return self

    def __next__(self):
        object = next(self.__iter)
        return object

    def search(self, pattern=None):
        if pattern:
            paginator = self.s3.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=self.s3_settings.BUCKET_NAME)
            return page_iterator.search(pattern)
        return iter(self)

    def clean(self, pattern=None):
        for object in self.search(pattern):
            self.delete(object["Key"])

    def get(self, key=None):
        try:
            return self.s3.get_object(
                Bucket=self.s3_settings.BUCKET_NAME, Key=key or self.data.file_key
            )
        except ClientError as ex:
            if ex.response["Error"]["Code"] == "NoSuchKey":
                return None
            raise ex

    def delete(self, key=None):
        return self.s3.delete_object(
            Bucket=self.s3_settings.BUCKET_NAME, Key=key or self.data.file_key
        )

    def __str__(self) -> str:
        return f"{self.__class__}: {self.data.file_key}"

    def __repr__(self) -> str:
        return f"{self.__class__}: {self.data.file_key}"


class FileInfoModel(BaseModel):
    id: str
    format: str
    prefix: Optional[str]


class SimpleS3Data:
    def __init__(
        self,
        content,
        /,
        file_key,
        mime="text/plain",
        file_info: FileInfoModel = None,
        parent=None,
    ):
        self.content = content
        self._file_key = file_key
        self.mime = mime
        self.parent = parent
        self.file_info = None

        if file_info:
            self.file_info = (
                file_info
                if isinstance(file_info, FileInfoModel)
                else FileInfoModel(**file_info)
            )

    @property
    def file_key(self):
        if self.file_info:
            sku = self.parent.sku if self.parent else ""
            prefix = self.file_info.prefix or ""
            return f"{prefix}.{sku}.{self.file_info.id}.{self.file_info.format.lower()}"
        else:
            return self._file_key



from typing import List

from pydantic import BaseModel

from gdshoplib.apps.platforms.base import Platform
from gdshoplib.packages.feed import Feed


class VKProductModel(BaseModel):
    ...


class VKManager(Platform, Feed):
    DESCRIPTION_TEMPLATE = "vk.txt"
    KEY = "VK"

    # def get_products(self) -> List[VKProductModel]:
    #     ...

    # def push_feed(self):
    #     # Обновить товары в VK
    #     ...



# Менеджер управления Notion
import functools

from gdshoplib.core.settings import NotionSettings
from gdshoplib.packages.cache import KeyDBCache
from gdshoplib.services.notion.manager import RequestManager


class Notion(RequestManager):
    def __init__(self, *args, caching=False, **kwargs) -> None:
        self.settings = NotionSettings()
        self.caching = caching
        self.cache = KeyDBCache()
        super(Notion, self).__init__(*args, **kwargs)

    def get_headers(self):
        return {
            **self.auth_headers(),
            "Content-Type": "application/json",
            "Notion-Version": "2022-06-28",
            "Accept": "application/json",
        }

    def cache_one(func):
        @functools.wraps(func)
        def wrap(self, id):
            if not self.settings.CACHE_ENABLED:
                return func(self, id)

            if self.caching or not self.cache.exists(id):
                self.cache[id] = func(self, id)

            return self.cache[id]

        return wrap

    def auth_headers(self):
        return {"Authorization": "Bearer " + self.settings.NOTION_SECRET_TOKEN}

    def get_user(self, user_id):
        return self.make_request(f"blocks/{user_id}", method="get")

    def get_capture(self, block):
        _capture = block[block["type"]].get("caption")
        return _capture[0].get("plain_text") if _capture else ""

    # TODO: Сделать декоратор пагинации
    def get_blocks(self, parent_id, params=None):
        blocks = []
        for block in self.pagination(
            f"blocks/{parent_id}/children", method="get", params=params
        ):
            if not block.get("has_children"):
                blocks.append(block)
            else:
                blocks.extend(self.get_blocks(block.get("id")))
        return blocks

    @cache_one
    def get_block(self, block_id):
        return self.make_request(f"blocks/{block_id}", method="get")

    @cache_one
    def get_page(self, page_id):
        return self.make_request(f"pages/{page_id}", method="get")

    @cache_one
    def get_database(self, database_id):
        return self.make_request(f"databases/{database_id}", method="get")

    def get_pages(self, database_id, params=None):
        return self.pagination(
            f"databases/{database_id}/query", method="post", params=params
        )

    def get_prop(self, product_id, param_id):
        self.make_request(f"pages/{product_id}/properties/{param_id}")

    def update_prop(self, product_id, params=None):
        # TODO: Переделать в обновление параметра как объекта
        self.make_request(
            f"pages/{product_id}",
            method="patch",
            params=params,
            # params={"properties": {"Наш SKU": [{"text": {"content": sku}}]}},
        )

    def update_block(self, block_id, params):
        self.make_request(
            f"blocks/{block_id}",
            method="patch",
            params=params
            # params={"code": {"rich_text": [{"text": {"content": content}}]}},
        )



from dateutil.parser import parse

from gdshoplib.services.notion.base import BasePage
from gdshoplib.services.notion.block import Block
from gdshoplib.services.notion.notion import Notion


class Page(BasePage):
    def refresh(self):
        Notion(caching=True).get_page(self.id)
        self.initialize()

    def initialize(self):
        super(Page, self).initialize()
        self.properties = PageProperty(self.page)

    def blocks(self, filter=None):
        if not filter:
            for block in self.notion.get_blocks(self.id):
                yield Block(block["id"], notion=self.notion, parent=self)
            return

        for block in self.notion.get_blocks(self.id):
            for k, v in filter.items():
                block = Block(block["id"], notion=self.notion, parent=self)
                if str(block.__getattr__(k)).lower() == str(v).lower():
                    yield block

    def commit(self):
        # Проитерироваться по изменениям и выполнить в Notion
        ...

    def to_json(self):
        # Вернуть товар в json
        ...


class PageProperty:
    def __init__(self, page):
        self.page = page

    def __getitem__(self, key):
        return self.__dict__.get(key) or self.__search_content(key)

    def __getattr__(self, key):
        return self[key]

    def __search_content(self, key):
        for _, prop in self.page.get("properties", {}).items():
            for prop_field in self.__get_prop(key):
                if prop["id"] == prop_field["id"]:
                    data = self.properties_type_parse_map.get(prop["type"])(prop)
                    handler = prop_field.get("handler")

                    if handler and data:
                        return handler(data)
                    return data
        raise AttributeError

    def __get_prop(self, key):
        return self.properties_keys_map[key]

    def __str__(self) -> str:
        return f"{self.__class__}"

    def __repr__(self) -> str:
        return f"{self.__class__}"

    def relation_handler(self, page_id_in_list):
        _id = page_id_in_list[0]["id"]
        return Page(_id)

    def relation_list_handler(self, page_ids):
        result = []
        for _id in page_ids:
            result.append(Page(_id["id"]))
        return result

    def date_handler(self, date):
        if not date:
            return
        elif "start" in date:
            return parse(date["start"]).date()

        return parse(date).date()

    def files_field_handler(self, files):
        if not files:
            return None
        if len(files) == 1:
            return files[0]["file"]["url"]
        return [file["file"]["url"] for file in files]

    @property
    def properties_keys_map(self):
        return {
            "title": (dict(name="Name", id="title"),),
            "edited_by": (dict(name="Last edited by", id="~%7BrF"),),
            "work": (dict(name="Использовать", id="bAcI"),),
            "created_time": (
                dict(name="Created time", id="v%5Dsj", handler=self.date_handler),
            ),
            "short_description": (dict(name="Короткое описание", id="u_tU"),),
            "size": (
                dict(name="Размер", id="taW%3B"),
                dict(name="Размер", id="NUuD"),
            ),
            "notes_field": (dict(name="Примечания", id="sXND"),),
            "quantity": (dict(name="Кол-во", id="pXTy"),),
            "edited_time": (
                dict(name="Last edited time", id="mVEw", handler=self.date_handler),
            ),
            "collection": (dict(name="Коллекция", id="W%5BhI"),),
            "name": (dict(name="Название на русском", id="Tss%5D"),),
            "created_by": (dict(name="Created by", id="TbyK"),),
            "kit_field": (
                dict(
                    name="Комплект", id="QV%5D%5D", handler=self.relation_list_handler
                ),
            ),
            "tags_field": (dict(name="Теги", id="MqdC"),),
            "status_description": (dict(name="Описание", id="MUl%7C"),),
            "status_publication": (dict(name="Публикация", id="BeEA"),),
            "color": (dict(name="Цвет", id="Jvku"),),
            "specifications_field": (
                dict(name="Материалы / Характеристики", id="COmf"),
            ),
            "price_neitral": (dict(name="Безубыточность", id="VmWm"),),
            "price_now": (dict(name="Текущая Цена", id="Ddaz"),),
            "price_kit": (dict(name="Цена комплекта", id="Dwfs"),),
            "price_current_discount": (dict(name="Текущая Скидка", id="syrp"),),
            "price_agent": (dict(name="Агентская Цена", id="vC%5E%3D"),),
            "price_agent_kit": (dict(name="Агентский комплект", id="M%60HY"),),
            "price_gross": (dict(name="Себестоимость", id="d%3DhO"),),
            "avito_id": (dict(name="", id="voQ%3E"),),
            "vk_id": (dict(name="", id="Ve%7Ca"),),
            "discount_from_date": (
                dict(name="Дата поставки", id="%60_a%3D", handler=self.date_handler),
            ),
            "badge_field": (
                dict(name="🎖️ Бейджи", id="%3CU_H", handler=self.relation_list_handler),
            ),
            "file": (dict(name="Файл", id="uK%3CN", handler=self.files_field_handler),),
            "coordinates": (dict(name="Координаты", id="XU%7C%3F"),),
            "transparency": (dict(name="Прозрачность", id="RHj%5B"),),
            "sku": (dict(name="Наш SKU", id="BKOs"),),
            "sku_s": (dict(name="SKU поставщика", id="BHve"),),
            "price_eur": (dict(name="Цена (eur)", id="AyqD"),),
            "platforms": (
                dict(name="Платформы", id="%40Q~A", handler=self.relation_list_handler),
            ),
            "brand": (
                dict(name="Бренд", id="gk%40%3B", handler=self.relation_handler),
            ),
            "categories": (
                dict(name="Категории", id="%7CFzB", handler=self.relation_list_handler),
            ),
            "price_coefficient": (
                dict(name="Коэфицент бренда", id="HjFs"),
                dict(name="Наценка", id="YsUp"),
                dict(name="Наценка", id="lNgd"),
            ),
        }

    @property
    def properties_type_parse_map(self):
        return {
            "rich_text": lambda data: " ".join(
                [t.get("plain_text", "") for t in data["rich_text"]]
            )
            or "",
            "number": lambda data: data["number"] or 0,
            "title": lambda data: data["title"][0]["text"]["content"],
            "select": lambda data: data.get("select").get("name")
            if data.get("select")
            else None,
            "multi_select": lambda data: data,
            "status": lambda data: data["status"]["name"],
            "date": lambda data: data["date"],
            "formula": lambda data: data["formula"]["number"],
            "relation": lambda data: data["relation"],
            "rollup": lambda data: data,
            "people": lambda data: data,
            "files": lambda data: data["files"],
            "checkbox": lambda data: data["checkbox"],
            "url": lambda data: data["url"],
            "email": lambda data: data,
            "phone_number": lambda data: data,
            "created_time": lambda data: parse(data["created_time"]),
            "created_by": lambda data: str(data["created_by"]),
            "last_edited_time": lambda data: parse(data["last_edited_time"]),
            "last_edited_by": lambda data: str(data["last_edited_by"]),
            "image": lambda data: data["image"]["file"]["url"],
            "video": lambda data: data["video"]["file"]["url"],
        }



from lxml import etree, objectify

from gdshoplib.apps.platforms.base import Platform
from gdshoplib.packages.feed import Feed


class AvitoManager(Platform, Feed):
    KEY = "AVITO"

    def get_root(self):
        root = etree.Element("Ads")
        objectify.deannotate(root, cleanup_namespaces=True, xsi_nil=True)
        root.attrib["target"] = "Avito"
        root.attrib["formatVersion"] = "3"
        return root

    def get_offer(self, product):
        appt = objectify.Element("Ad")
        appt.Id = product.sku
        appt.Title = product.name
        appt.Category = self.feed_settings.AVITO_CATEGORY
        appt.Condition = "Новое"
        appt.Description = product.description.render(self)
        appt.Address = self.feed_settings.ADDRESS
        appt.Price = product.price.now
        appt.ManagerName = self.feed_settings.MANAGER_NAME
        appt.ContactPhone = self.feed_settings.PHONE

        images = objectify.Element("Images")
        for image in product.images:
            i = objectify.Element("Image")
            i.attrib["url"] = self.get_media_url(image)
            images.append(i)

        appt.append(images)

        objectify.deannotate(appt, cleanup_namespaces=True, xsi_nil=True)
        return appt

    def get_feed(self):
        for product in self.products:
            self.root.append(self.get_offer(product))

        return etree.tostring(
            self.root,
            pretty_print=True,
            encoding="utf-8",
            xml_declaration=True,
            standalone=True,
        )
